﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using Kbg.NppPluginNET.PluginInfrastructure;
using System.Text;

namespace NppLinefilter {
    internal class LineFilter {

        private enum BufferEncoding { Ansi = 0, Utf8bom = 1, ucs2be = 2, ucs2le = 3, Utf8 = 4 }
        // enum UniMode - uni8Bit 0, uniUTF8 1, uni16BE 2, uni16LE 3, uniCookie 4, uni7Bit 5, uni16BE_NoBOM 6, uni16LE_NoBOM 7

        private enum LineBreak { Unbestimmt, Windows, Unix, Macintosh }

        /**
         * Text Search
         * */

        internal static void KeepLinesWithSelection(bool ignoreCase) {
            string selection = GetSelection();
            ProcessLinesWith(selection, ignoreCase, true);
        }

        internal static void RemoveLinesWithSelection(bool ignoreCase) {
            string selection = GetSelection();
            ProcessLinesWith(selection, ignoreCase, false);
        }

        internal static void KeepLinesWith(string text, bool ignoreCase) {
            ProcessLinesWith(text, ignoreCase, true);
        }

        internal static void RemoveLinesWith(string text, bool ignoreCase) {
            ProcessLinesWith(text, ignoreCase, false);
        }

        private static void ProcessLinesWith(string text, bool ignoreCase, bool shouldContainSelection) {
            Func<string, bool> keepLineMethod = (string line) => {
                return shouldContainSelection == LineHasSelection(line, text, ignoreCase);
            };

            ProcessContent(keepLineMethod);
        }

        private static string GetSelection() {
            return CreateGateway().GetSelText();
        }

        private static bool LineHasSelection(string line, string selection, bool ignoreCase) {
            return line.IndexOf(selection, ignoreCase ? StringComparison.OrdinalIgnoreCase : StringComparison.Ordinal) >= 0;
        }


        /**
         * Regex Search
         * */

        internal static void KeepLinesWithRegex(string regex) {
            ProcessLinesWithRegex(regex, true);
        }

        internal static void RemoveLinesWithRegex(string regex) {
            ProcessLinesWithRegex(regex, false);
        }

        private static void ProcessLinesWithRegex(string regex, bool shouldMatch) {
            Func<string, bool> keepLineMethod = (string line) => {
                return shouldMatch == Regex.IsMatch(line, regex);
            };

            ProcessContent(keepLineMethod);
        }


        /**
         * Common
         * */

        private static ScintillaGateway CreateGateway() {
            Encoding encoding = GetEncoding() == (int)BufferEncoding.Ansi ? Encoding.Default : Encoding.UTF8;

            return new ScintillaGateway(PluginBase.GetCurrentScintilla(), encoding);
        }


        private static int GetEncoding() {
            IntPtr currentBuffer = Win32.SendMessage(PluginBase.nppData._nppHandle, (uint)NppMsg.NPPM_GETCURRENTBUFFERID, 0, 0);
            return (int)Win32.SendMessage(PluginBase.nppData._nppHandle, (uint)NppMsg.NPPM_GETBUFFERENCODING, currentBuffer, 0);
        }


        private static void ProcessContent(Func<string, bool> KeepLineMethod) {
            try {
                List<string> newFileContent = new List<string>();

                ScintillaGateway scintilla = CreateGateway();

                int numLines = scintilla.GetLineCount();

                LineBreak lineBreak = LineBreak.Unbestimmt;

                for (int i = 0; i < numLines; i++) {
                    string line = scintilla.GetLine(i);

                    if (lineBreak == LineBreak.Unbestimmt) {
                        lineBreak = BestimmeLineBreak(line);
                    }

                    if (KeepLineMethod(line)) {
                        newFileContent.Add(line);
                    }
                }

                int encoding = GetEncoding();

                Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FILE_NEW);

                if (encoding == (int)BufferEncoding.Ansi) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_ANSI);
                } else if (encoding == (int)BufferEncoding.Utf8) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_AS_UTF_8);
                } else if (encoding == (int)BufferEncoding.Utf8bom) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_UTF_8);
                } else if (encoding == (int)BufferEncoding.ucs2be) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_UCS_2BE);
                } else if (encoding == (int)BufferEncoding.ucs2le) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_UCS_2LE);
                }

                if (lineBreak == LineBreak.Macintosh) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_TOMAC);
                } else if (lineBreak == LineBreak.Unix) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_TOUNIX);
                } else if (lineBreak == LineBreak.Windows) {
                    Win32.SendMessage(PluginBase.nppData._nppHandle, (uint) NppMsg.NPPM_MENUCOMMAND, 0, NppMenuCmd.IDM_FORMAT_TODOS);
                }

                newFileContent.ForEach(line => {
                    scintilla.AppendText(line);
                });
            } catch (Exception e) {
                // Debugging-Variante:
                //MessageBox.Show("Sorry an unknown exception occured during filtering.\r\n" + e.Message + "\r\n" + e.StackTrace, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                MessageBox.Show("Sorry an unknown exception occured during filtering.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private static LineBreak BestimmeLineBreak(string line) {
            // 0x13 & 0x10  -->  \r\n  -->  Windows
            // 0x13         -->  \r    -->  Macintosh
            // 0x10         -->  \n    -->  Unix
            if (line.Length == 0) {
                return LineBreak.Unbestimmt;
            }

            if (line.Length == 1) {
                byte lastByte = (byte)line.ToCharArray(line.Length - 1, 1)[0];

                return BestimmeLineBreakFromLastByte(lastByte);
            }

            char[] lastTwoChars = line.ToCharArray(line.Length - 2, 2);
            byte byte0 = (byte)lastTwoChars[0];
            byte byte1 = (byte)lastTwoChars[1];

            return byte0 == 13 && byte1 == 10 ? LineBreak.Windows : BestimmeLineBreakFromLastByte(byte1);
        }

        private static LineBreak BestimmeLineBreakFromLastByte(byte lastByte) {
            if (lastByte == 13) {
                return LineBreak.Macintosh;
            } if (lastByte == 10) {
                return LineBreak.Unix;
            } else {
                return LineBreak.Unbestimmt;
            }
        }
    }
}